/******************************************************************************
 * Copyright(c) 2008 - 2010 Realtek Corporation. All rights reserved.
 *
 * Based on the r8180 driver, which is:
 * Copyright 2004-2005 Andrea Merello <andreamrl@tiscali.it>, et al.
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 * The full GNU General Public License is included in this distribution in the
 * file called LICENSE.
 *
 * Contact Information:
 * Jerry chuang <wlanfae@realtek.com>
******************************************************************************/

#include "r8180.h"
#include "r8180_hw.h"
#include "r8180_sa2400.h"



u32 sa2400_chan[] = {
	0x0,	
	0x00096c, 
	0x080970, 
	0x100974, 
	0x180978, 
	0x000980, 
	0x080984, 
	0x100988, 
	0x18098c, 
	0x000994, 
	0x080998, 
	0x10099c, 
	0x1809a0, 
	0x0009a8, 
	0x0009b4, 
};
	

void rf_stabilize(struct net_device *dev)
{
	force_pci_posting(dev);
	mdelay(3); 
}


void write_sa2400(struct net_device *dev,u8 adr, u32 data) 
{
	u32 phy_config;


	
	phy_config = 0xb0000000; 

	phy_config |= (((u32)(adr&0xf))<< 24);
	phy_config |= (data & 0xffffff);
	write_nic_dword(dev,PHY_CONFIG,phy_config);
#ifdef DEBUG_SA2400
	DMESG("Writing sa2400: %x (adr %x)",phy_config,adr);
#endif
	rf_stabilize(dev);
}



void sa2400_write_phy_antenna(struct net_device *dev,short ch)
{
	struct r8180_priv *priv = ieee80211_priv(dev);
	u8 ant;

	ant = SA2400_ANTENNA;
	if(priv->antb) 
		ant |= BB_ANTENNA_B;
	if(ch == 14)
		ant |= BB_ANTATTEN_CHAN14;
	write_phy(dev,0x10,ant);
}


short sa2400_rf_set_sens(struct net_device *dev, short sens)
{
	u8 finetune = 0;
	if ((sens > 85) || (sens < 54)) return -1;
	
	write_sa2400(dev,5,0x1dfb | (sens-54) << 15 |(finetune<<20));  
	
	return 0;
}


void sa2400_rf_set_chan(struct net_device *dev, short ch)
{
	struct r8180_priv *priv = ieee80211_priv(dev);
	u32 txpw = 0xff & priv->chtxpwr[ch]; 
	u32 chan = sa2400_chan[ch];
	
	write_sa2400(dev,7,txpw);
	sa2400_write_phy_antenna(dev,ch);
	write_sa2400(dev,0,chan);
	write_sa2400(dev,1,0xbb50);
	write_sa2400(dev,2,0x80);
	write_sa2400(dev,3,0);
}


void sa2400_rf_close(struct net_device *dev)
{
	write_sa2400(dev, 4, 0);
}


void sa2400_rf_init(struct net_device *dev)
{
	struct r8180_priv *priv = ieee80211_priv(dev);
	u32 anaparam;
	u8 firdac;
	
	write_nic_byte(dev,PHY_DELAY,0x6);	
	write_nic_byte(dev,CARRIER_SENSE_COUNTER,0x4c); 
	
	anaparam = read_nic_dword(dev,ANAPARAM);
	anaparam = anaparam &~ (1<<ANAPARAM_TXDACOFF_SHIFT);
	
	anaparam = anaparam &~ANAPARAM_PWR1_MASK;
	anaparam = anaparam &~ANAPARAM_PWR0_MASK;
	if(priv->digphy){
		anaparam |= (SA2400_DIG_ANAPARAM_PWR1_ON<<ANAPARAM_PWR1_SHIFT);
		anaparam |= (SA2400_ANAPARAM_PWR0_ON<<ANAPARAM_PWR0_SHIFT);
	}else{
		anaparam |= (SA2400_ANA_ANAPARAM_PWR1_ON<<ANAPARAM_PWR1_SHIFT);
	}
	
	rtl8180_set_anaparam(dev,anaparam);
	
	firdac = (priv->digphy) ? (1<<SA2400_REG4_FIRDAC_SHIFT) : 0;
	write_sa2400(dev,0,sa2400_chan[priv->chan]); 
	write_sa2400(dev,1,0xbb50);
	write_sa2400(dev,2,0x80);
	write_sa2400(dev,3,0);
	write_sa2400(dev,4,0x19340 | firdac); 
	write_sa2400(dev,5,0xc9dfb);  
	write_sa2400(dev,4,0x19348 | firdac);  
	
	if(priv->digphy)
		write_sa2400(dev,4,0x1938c); 
		
	write_sa2400(dev,4,0x19340 | firdac);
	
	write_sa2400(dev,0,sa2400_chan[priv->chan]); 
	write_sa2400(dev,1,0xbb50);
	write_sa2400(dev,2,0x80);
	write_sa2400(dev,3,0);
	write_sa2400(dev,4,0x19344 | firdac); 
	
	write_sa2400(dev,6,0x13ff | (1<<23)); 
	write_sa2400(dev,8,0); 
	
	if(!priv->digphy)
	{
		rtl8180_set_anaparam(dev, anaparam | \
				     (1<<ANAPARAM_TXDACOFF_SHIFT));
		
		rtl8180_conttx_enable(dev);
		
		write_sa2400(dev, 4, 0x19341); 

		write_sa2400(dev, 4, 0x19345);
		rtl8180_conttx_disable(dev);
		
		rtl8180_set_anaparam(dev, anaparam);
	}
	
	write_sa2400(dev,4,0x19341 | firdac ); 
	
	

	write_phy(dev,0,0x98);
	write_phy(dev,3,0x38);
	write_phy(dev,4,0xe0);
	write_phy(dev,5,0x90);
	write_phy(dev,6,0x1a);
	write_phy(dev,7,0x64);
	
	
	sa2400_write_phy_antenna(dev,priv->chan);
	
	write_phy(dev,0x11,0x80);
	if(priv->diversity)
		write_phy(dev,0x12,0xc7);
	else
		write_phy(dev,0x12,0x47); 
	
	write_phy(dev,0x13,0x90 | priv->cs_treshold ); 
	
	write_phy(dev,0x19,0x0);
	write_phy(dev,0x1a,0xa0);
	
	sa2400_rf_set_chan(dev,priv->chan);
}
